# -*- test-case-name: mimir.monitor.test.test_news -*-
#
# Copyright (c) 2005-2008 Ralph Meijer
# See LICENSE for details

import re, time

import feedparser, simplejson

from zope.interface import Interface, implements

from twisted.application import service
from twisted.internet import reactor
from twisted.python import components, failure
from twisted.words.xish import domish

from wokkel import pubsub
from wokkel.iwokkel import IXMPPHandler

SGMLTAG = re.compile('<.+?>', re.DOTALL)
NS_ATOM = 'http://www.w3.org/2005/Atom'

class FeedParserEncoder(simplejson.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, time.struct_time):
            return tuple(obj)
        elif isinstance(obj, Exception):
            return failure.Failure(obj).getTraceback()
        else:
            return simplejson.JSONEncoder.default(self, obj)

class NoNotify(Exception):
    pass

class INewsService(Interface):

    def process(channel, items):
        """
        Process new items for channel

        @param items: atom entries
        @type items: L{domish.Element}
        """

class NewsService(service.Service):

    implements(INewsService)

    def __init__(self, presenceMonitor, dbpool):
        self._dbpool = dbpool
        presenceMonitor.register_callback(self.onPresenceChange)

    def error(self, failure):
        print failure

    def onPresenceChange(self, entity, available, show):
        if available:
            if show not in ['away', 'xa', 'dnd', 'chat']:
                show = 'online'
        else:
            show = 'offline'

        print "  presence change to %s for %r", (show, entity.full())
        reactor.callLater(5, self.pageNotify, entity, show)

    def pageNotify(self, entity, show):
        d = self._dbpool.runInteraction(self._checkNotify, entity, show)
        d.addCallback(self._doNotify, entity)
        d.addCallback(self._setNotified)
        d.addErrback(lambda failure: failure.trap(NoNotify))
        d.addErrback(self.error)

    def _checkNotify(self, cursor, entity, show):
        cursor.execute("""SELECT user_id, message_type, ssl,
                                 count(news_id) as count
                          FROM auth_user
                          NATURAL JOIN news_prefs
                          NATURAL JOIN news_notify_presences
                          NATURAL JOIN news_page
                          NATURAL JOIN news_flags
                          NATURAL JOIN news
                          WHERE jid=%s AND
                                NOT suspended AND
                                presence=%s AND
                                NOT notified AND
                                date>last_visit
                          GROUP BY user_id, message_type, ssl""",
                       (entity.userhost(),
                        show))
        result = cursor.fetchone()

        if not result:
            return failure.Failure(NoNotify())

        return result

    def _doNotify(self, result, entity):
        userId, messageType, ssl, count = result

        title = u'New news on Mim\xedr!'
        link = "%s://mimir.ik.nu/news" % (ssl and 'https' or 'http')
        description = 'There '
        if count == 1:
            description += 'is 1 new item'
        else:
            description += 'are %s new items' % count
        description += ' on your news page'

        self.notifier.sendNotification(entity.full(), True, messageType,
                                       title, link, description)
        return userId

    def _setNotified(self, userId):
        return self._dbpool.runOperation("""UPDATE news_page SET notified=true
                                            WHERE user_id=%s""", userId)

    def process(self, channel, items):
        print "Got entries: %r" % items
        d = self._dbpool.runInteraction(self._processItems, channel, items)
        d.addCallback(self.notify)
        d.addErrback(self.error)

    def _processItems(self, cursor, channel, items):

        # Get channel title
        cursor.execute("SELECT title from channels WHERE channel=%s",
                       channel)
        title = cursor.fetchone()[0]

        print "Channel title: %r" % title

        feedDocument = domish.Element((NS_ATOM, 'feed'))
        for item in items:
            feedDocument.addChild(item)

        feed = feedparser.parse(feedDocument.toXml().encode('utf-8'))
        entries = feed.entries

        # Get notify list, including preferences
        cursor.execute("""SELECT user_id, jid, notify,
                                 description_in_notify, message_type,
                                 store_offline, notify_items
                          FROM news_prefs
                            NATURAL JOIN news_subscriptions
                            NATURAL JOIN news_notify
                          WHERE NOT suspended AND channel=%s""",
                       channel)

        notifyList = cursor.fetchall()

        # split notify list into a list of entities that are to be notified
        # and a list of entities that will get items marked as unread
        notifications = []
        markUnread = []
        for userId, jid, notify, descriptionInNotify, messageType, \
            storeOffline, notifyItems in \
                notifyList:
            if notify and notifyItems:
                notifications.append((jid,
                                      descriptionInNotify,
                                      messageType))
            elif storeOffline:
                markUnread.append(userId)

        # store items and mark unread
        notifyItems = []
        for entry in entries:
            newsId = self._storeItem(cursor, channel, entry)
            if newsId:
                for userId in markUnread:
                    cursor.execute("""INSERT INTO news_flags
                                      (user_id, news_id, unread) VALUES
                                      (%s, %s, true)""",
                                   (userId, newsId))
                    cursor.execute("""UPDATE news_page SET notified=false
                                      WHERE user_id=%s""",
                                   userId)

                notifyItems.append(entry)
        return (title, notifications, notifyItems)

    def _extractBasics(self, entry):
        if 'title' in entry:
            title = entry.title

            if 'source' in entry and 'title' in entry.source:
                content = "%s: %s" % (entry.source.title, title)

            if entry.title_detail == 'text/plain':
                title = feedparser._xmlescape(title)
        else:
            title = u''

        if 'link' in entry:
            link = entry.get('feedburner_origlink', entry.link)
        else:
            link = u''

        # Find a description. First try full text, then summary.
        content = None
        if 'content' in entry and entry.content[0].value:
            content = entry.content[0]
        elif 'summary' in entry:
            content = entry.summary_detail

        if content:
            value = content.value
            if content.type == 'text/plain':
                value = feedparser._xmlescape(value)
            description = value
        else:
            description = u''

        date = None
        for attribute in 'updated', 'published', 'created':
            if attribute in entry:
                date = getattr(entry, '%s_parsed' % attribute)

        date = time.strftime("%Y-%m-%d %H:%M:%Sz", date or time.gmtime())

        return title, link, description, date

    def _storeItem(self, cursor, channel, entry):
        title, link, description, date = self._extractBasics(entry)
        json = simplejson.dumps(entry, cls=FeedParserEncoder)

        print "Storing item: %r" % entry.id

        cursor.execute("""UPDATE news
                          SET title=%s, description=%s, date=%s, parsed=%s
                          WHERE channel=%s AND link=%s""",
                       (title, description, date, json, channel, link))

        if cursor.rowcount == 1:
            print "UPDATE"
            return None

        cursor.execute("""INSERT INTO news
                          (channel, title, link, description, date, parsed)
                          VALUES
                          (%s, %s, %s, %s, %s, %s)""",
                       (channel, title, link, description, date, json))

        print "INSERT",

        cursor.execute("""SELECT news_id FROM news
                          WHERE channel=%s and link=%s""",
                          (channel, link))

        newsId = cursor.fetchone()[0]

        print newsId
        return newsId

    def notify(self, result):
        channelTitle, notifications, entries = result

        for entry in entries:
            title = "%s: %s" % (channelTitle,
                                entry.get('title', u'-- no title --'))
            link = entry.get('link', '')
            description = entry.get('summary')

            if description:
                description = SGMLTAG.sub('', description)
                description = domish.unescapeFromXml(description)
                description = description.rstrip() or None

            for jid, descriptionInNotify, messageType in notifications:
                self.notifier.sendNotification(jid, descriptionInNotify,
                                               messageType, title, link,
                                               description)

class XMPPHandlerFromService(pubsub.PubSubClient):

    def __init__(self, service):
        self.service = service

    def sendNotification(self, jid, descriptionInNotify, messageType,
                         title, link, description):

        message = domish.Element((None, 'message'))
        message['to'] = jid
        message['type'] = messageType

        if messageType == 'chat':
            body = "%s\n%s" % (title, link)
            if description and descriptionInNotify:
                body += "\n\n%s\n\n" % description

            message.addElement('body', None, body)
        elif messageType == 'headline':
            message.addElement('subject', None, title)
            if description:
                message.addElement('body', None, description)
            oob = message.addElement(('jabber:x:oob', 'x'))
            oob.addElement('url', None, link)
            oob.addElement('desc', None, title)

        print "Sending: %r" % message.toXml()
        self.send(message)

    def itemsReceived(self, event):
        m = re.match(r"^mimir/news/(.+)$", event.nodeIdentifier)

        if not m:
            return

        channel = m.group(1)

        entries = (item.entry for item in event.items
                              if item.entry and item.entry.uri == NS_ATOM)

        self.service.process(channel, entries)

components.registerAdapter(XMPPHandlerFromService,
                           INewsService,
                           IXMPPHandler)



# Copyright (c) 2005-2008 Ralph Meijer
# See LICENSE for details

from twisted.internet import defer
from twisted.trial import unittest
from twisted.words.protocols.jabber import xmlstream

from mimir.aggregator import aggregator

class DummyManager(object):
    def __init__(self):
        self.outlist = []

    def send(self, obj):
        self.outlist.append(obj)


class DummyAggregator(object):
    def __init__(self):
        self.feeds = {}

    def setFeed(self, handle, url):
        self.feeds[handle] = url
        return defer.succeed({'handle': handle, 'url': url})


class XMPPControlTest(unittest.TestCase):
    def test_onFeed(self):
        """
        Test adding a feed through XMPP.
        """
        xc = aggregator.XMPPControl(DummyAggregator())
        xc.parent = DummyManager()

        iq = xmlstream.IQ(None)
        iq['to'] = 'user1@example.org'
        iq['from'] = 'user2@example.org'
        iq.addElement(('http://mimir.ik.nu/protocol/aggregator', 'aggregator'))
        feed = iq.aggregator.addElement('feed') 
        feed.addElement('handle', content='test')
        feed.addElement('url', content='http://www.example.org/')
        xc.onFeed(iq)
        self.assertEquals(xc.parent.outlist[0]['type'], 'result')
        self.assertEquals(xc.service.feeds['test'], 'http://www.example.org/')


class MemoryFeedStorage(object):
    def __init__(self):
        self.feeds = {}
        self.feedList = {}

    def getFeedList(self):
        return defer.succeed(self.feedList)

    def setFeedURL(self, handle, url):
        feed = {'handle': handle, 'href': url}
        self.feedList[handle] = feed
        return defer.succeed(feed)

    def getFeed(self, handle):
        return defer.succeed(self.feeds[handle])

    def storeFeed(self, feed):
        self.feeds[feed['handle']] = feed
        return defer.succeed(None)

class AggregatorTest(unittest.TestCase):

    def test_setFeed(self):
        """
        Test setting a feed to be aggregated.

        The feed should have been added to storage, and aggregation
        scheduled.
        """

        feed = {'handle': 'test', 'href': 'http://example.org/feed/atom'}

        rescheduled = []

        def reschedule(delay, handle, useCache):
            rescheduled.append((delay, handle, useCache))

        def cb2(feedList):
            self.assertIn('test', feedList)
            self.assertEquals(feed, feedList['test'])

        def cb(result):
            self.assertEquals([(0, 'test', 0)], rescheduled)
            self.assertEquals(feed, result)
            d = storage.getFeedList()
            d.addCallback(cb2)
            return d

        storage = MemoryFeedStorage()
        agg = aggregator.AggregatorService(storage)
        agg.reschedule = reschedule
        agg.startService()

        d = agg.setFeed('test', 'http://example.org/feed/atom')
        d.addCallback(cb)
        return d

    def test_setFeedInvalidHandle(self):
        """
        Test setting a feed to be aggregated, with an invalid handle.

        The feed should not have been added to storage, and aggregation
        not scheduled.
        """

        rescheduled = []

        def reschedule(delay, handle, useCache):
            rescheduled.append((delay, handle, useCache))

        def cb(result):
            self.assertEquals([], rescheduled)

        storage = MemoryFeedStorage()
        agg = aggregator.AggregatorService(storage)
        agg.reschedule = reschedule
        agg.startService()

        d = agg.setFeed('!test', 'http://example.org/feed/atom')
        self.assertFailure(d, aggregator.InvalidHandleError)
        d.addCallback(cb)
        return d



# Copyright (c) 2005-2006 Ralph Meijer
# See LICENSE for details

"""
Unparsers for feeds.

Writers create another representation from the Universal Feed Parser's entry
representation.
"""

from twisted.words.xish import domish

NS_XML = 'http://www.w3.org/XML/1998/namespace'
NS_ATOM = 'http://www.w3.org/2005/Atom'

class AtomWriter(object):
    """
    Writer that generates entries that adhere to the Atom specification.
    """

    def generate(self, feed, entry):
        """
        Generate an Atom entry from a feedparser entry.

        @param entry: a parsed feed entry.
        @type entry: L{dict}
        @return: an Atom entry.
        @rtype: L{domish.Element}
        """

        element = domish.Element((NS_ATOM, 'entry'))

        for key, value in entry.iteritems():
            methodname = "_generate_" + key
            try:
                method = getattr(self, methodname)
            except AttributeError:
                continue

            result = method(value)

            if not isinstance(result, list):
                result = [result]

            for item in result:
                element.addChild(item)

        return element

    def generate_person(self, name, data):
        element = domish.Element((None, name))

        for key, value in data.iteritems():
            if key in ['name', 'uri', 'email']:
                element.addElement(key, content=value)

        return element

    def generate_element(self, name, data):
        element = domish.Element((None, name))

        if data:
            element.addContent(data)

        return element

    def generate_text(self, name, data):
        element = domish.Element((None, name))

        if data['value']:
            if data.type == "text/plain":
                element['type'] = 'text'
                element.addContent(domish.escapeToXml(data.value))
            else:
                element['type'] = 'html'
                element.addContent(data.value)

            if 'language' in data and data['language']:
                element[(NS_XML, 'lang')] = data['language']

            if 'base' in data:
                element[(NS_XML, 'base')] = data['base']

        return element

    def _generate_author_detail(self, data):
        return self.generate_person('author', data)

    def _generate_contributor(self, data):
        return [self.generate_person('contributor', item)
                for item in data]

    def _generate_content(self, data):
        elements = []
        for item in data:
            element = self.generate_text('content', item)

            if 'src' in item:
                element['src'] = item['src']

            elements.append(element)

        return elements

    def _generate_feedburner_origlink(self, data):
        element = domish.Element(('http://rssnamespace.org/feedburner/ext/1.0',
                                  'origLink'))
        element.addContent(data)
        return element

    def _generate_id(self, data):
        return self.generate_element('id', data)

    def _generate_links(self, data):
        elements = []
        for item in data:
            element = domish.Element((None, 'link'))
            for key, value in item.iteritems():
                if key in ['rel', 'type', 'href', 'hreflang', 'title', 'length']:
                    element[key] = value

            elements.append(element)

        return elements

    def _generate_published(self, data):
        return self.generate_element('published', data)

    def _generate_subtitle_detail(self, data):
        return self.generate_text('subtitle', data)

    def _generate_summary_detail(self, data):
        return self.generate_text('summary', data)

    def _generate_tags(self, data):
        elements = []

        for item in data:
            element = domish.Element((None, 'category'))
            for key, value in item.iteritems():
                if value and key in ['term', 'scheme', 'label']:
                    element[key] = value

            elements.append(element)

        return elements

    def _generate_title_detail(self, data):
        return self.generate_text('title', data)

    def _generate_updated(self, data):
        return self.generate_element('updated', data)

class MimirWriter(object):
    """
    Writer that generates entries using a custom XML representation.
    """

    def generate(self, feed, entry):
        news = domish.Element(('mimir:news', 'news'))

        if 'title' in entry:
            content = entry.title
            if 'source' in entry and 'title' in entry.source:
                content = "%s: %s" % (entry.source.title, content)
            if entry.title_detail.type == 'text/plain':
                content = domish.escapeToXml(content)
            news.addElement('title', content=content)

        if 'link' in entry:
            url = entry.get('feedburner_origlink', entry.link)

            if url:
                news.addElement('link', content=url)

        # Find a description. First try full text, then summary.
        content = None
        if 'content' in entry:
            content = entry.content[0]
        elif 'summary' in entry:
            content = entry.summary_detail

        if content:
            value = content.value
            if content.type == 'text/plain':
                value = domish.escapeToXml(value)
            news.addElement('description', content=value)

        return news

class StringParser(object):
    """
    Parses a string with serialized XML into a DOM.
    """
    def __init__(self):
        self.elementStream = domish.elementStream()
        self.elementStream.DocumentStartEvent = self.docStart
        self.elementStream.ElementEvent = self.elem
        self.elementStream.DocumentEndEvent = self.docEnd
        self.done = False


    def docStart(self, elem):
        self.document = elem


    def elem(self, elem):
        self.document.addChild(elem)


    def docEnd(self):
        self.done = True


    def parse(self, string):
        self.elementStream.parse(string)
        if not self.done:
            raise Exception("Incomplete XML document input")
        return self.document



class ReconstituteWriter(object):
    """
    Writer that generates Atom entries using Venus' reconsitute module.
    """

    def generate(self, feed, entry):
        from planet.reconstitute import reconstitute

        xdoc = reconstitute(feed, entry)
        xml = xdoc.documentElement.toxml()
        return StringParser().parse(xml.encode('utf-8'))



# Copyright (c) 2005-2006 Ralph Meijer
# See LICENSE for details

"""
Create a monitor service.
"""

from twisted.application import service
from twisted.enterprise import adbapi
from twisted.python import usage
from twisted.words.protocols.jabber import jid

from wokkel import client
from wokkel.iwokkel import IXMPPHandler

from mimir.monitor import news, presence

class Options(usage.Options):
    optParameters = [
        ('jid', None, None),
        ('secret', None, None),
        ('dbuser', None, None),
        ('dbname', None, 'mimir'),
    ]

    optFlags = [
        ('verbose', 'v', 'Show traffic'),
    ]

    def postOptions(self):
        try:
            self['jid'] = jid.JID(self['jid'])
        except jid.InvalidFormat:
            raise usage.UsageError("'%(jid)s' is not a valid Jabber ID" % self)

        if not self['secret']:
            raise usage.UsageError("No secret provided")

def makeService(config):
    s = service.MultiService()

    clientService = client.XMPPClient(config['jid'], config['secret'])
    clientService.setServiceParent(s)

    clientService.factory.maxDelay = 900

    if config["verbose"]:
        clientService.logTraffic = True

    dbpool = adbapi.ConnectionPool('pyPgSQL.PgSQL',
                                   user=config["dbuser"],
                                   database=config["dbname"],
                                   client_encoding='utf-8',
                                   cp_min = 1,
                                   cp_max = 1,
                                   cp_reconnect=True,
                                   )

    ms = presence.Storage(dbpool)
    presenceMonitor = presence.RosterMonitor(ms)
    presenceMonitor.setHandlerParent(clientService)

    newsService = news.NewsService(presenceMonitor, dbpool)
    newsService.setServiceParent(s)

    xep = IXMPPHandler(newsService)
    newsService.notifier = xep
    xep.setHandlerParent(clientService)

    return s



# Copyright (c) 2005-2007 Ralph Meijer
# See LICENSE for details

from wokkel.xmppim import PresenceClientProtocol

class Storage(object):
    def __init__(self, dbpool):
        self._dbpool = dbpool
        d = self._dbpool.runOperation("""UPDATE presences
                                         SET type='unavailable', show='',
                                             status='', priority=0
                                         WHERE type='available'""")
        def eb(failure):
            print failure
        d.addErrback(eb)

    def set_presence(self, entity, available, show, status, priority):
        return self._dbpool.runInteraction(self._set_presence, entity,
                                                               available,
                                                               show,
                                                               status,
                                                               priority)

    def _set_presence(self, cursor, entity, available, show, status, priority):
        if available:
            type = 'available'
        else:
            type = 'unavailable'

        show = show or ''
        status = status or ''

        # changed is True when this resource became the top resource, or when
        # it continued to be the top resource and the availability or show
        # changed, or when another resource became the top resource
        changed = False

        # Find existing entry for this resource
        cursor.execute("""SELECT presence_id, type, show FROM presences
                          WHERE jid=%s AND resource=%s""",
                       (entity.userhost(), entity.resource))
        result = cursor.fetchone()
        print "result: %r" % result

        if result:
            id, old_type, old_show = result

            if old_type == 'unavailable':
                # delete old record, the new record will be inserted below
                cursor.execute("DELETE FROM presences WHERE presence_id=%s",
                               id)

        if result and old_type == 'available':
            if show != old_show:
                print "  show != old_show"
                changed = True
            cursor.execute("""UPDATE presences SET
                              type=%s, show=%s, status=%s, priority=%s,
                              last_updated=now()
                              WHERE presence_id=%s""",
                           (type, show, status, priority, id))
        else:
            print "  new presence record"
            changed = True
            cursor.execute("""INSERT INTO presences
                              (type, show, status, priority, jid, resource)
                              VALUES (%s, %s, %s, %s, %s, %s)""",
                           (type, show, status, priority,
                            entity.userhost(), entity.resource))


        return changed

    def update_roster(self, changed, entity):
        return self._dbpool.runInteraction(self._update_roster, changed,
                                                                entity)

    def _update_roster(self, cursor, changed, entity):
        print "Updating roster for %r" % entity.full()

        # Find new top resource's presence id
        cursor.execute("""SELECT presence_id, resource FROM presences
                          WHERE jid=%s ORDER by type, priority desc,
                          (CASE WHEN type='available'
                                THEN presence_id
                                ELSE 0
                           END), last_updated desc""",
                           entity.userhost())
        result = cursor.fetchone()
        top_id, top_resource = result

        # Get old top resource's presence id.
        cursor.execute("SELECT presence_id FROM roster WHERE jid=%s",
                                       entity.userhost())
        result = cursor.fetchone()
        print "result 2: %r" % result

        if result:
            old_top_id = result[0]
            print "  old_top_id %d" % old_top_id

            if old_top_id != top_id:
                print "  old_top_id != top_id"
                changed = True
            elif entity.resource != top_resource:
                print "  we are not the top resource"
                changed = False
            # else, we are still the top resource. Keep the changed value
            # that got passed.

            cursor.execute("UPDATE roster SET presence_id=%s WHERE jid=%s",
                           (top_id, entity.userhost()))
        else:
            changed = True
            cursor.execute("""INSERT INTO roster
                              (presence_id, jid) VALUES
                              (%s, %s)""",
                           (top_id, entity.userhost()))

        return changed

    def remove_presences(self, entity):
        return self._dbpool.runInteraction(self._remove_presences, entity)

    def _remove_presences(self, cursor, entity):
        cursor.execute("DELETE FROM roster WHERE jid=%s", entity.userhost())
        cursor.execute("DELETE FROM presences WHERE jid=%s", entity.userhost())

class Monitor(PresenceClientProtocol):
    def __init__(self, storage):
        self.storage = storage
        self.callbacks = []

    def connectionInitialized(self):
        PresenceClientProtocol.connectionInitialized(self)
        self.available()

    def register_callback(self, f):
        self.callbacks.append(f)

    def store_presence(self, entity, available, show, status, priority):
        d = self.storage.set_presence(entity, available, show, status, priority)
        d.addCallback(self.storage.update_roster, entity)
        def cb(changed, entity):
            print "Changed %r: %s" % (entity.full(), changed)
            if changed:
                for f in self.callbacks:
                    f(entity, available, show)

        d.addCallback(cb, entity)
        d.addErrback(self.error)

    def availableReceived(self, entity, show, statuses, priority):
        print "available: %r" % entity.full()
        if statuses:
            status = statuses.popitem()[1]
        else:
            status = None

        print "  status: %r" % status
        self.store_presence(entity, True, show, status, priority)

    def unavailableReceived(self, entity, statuses):
        if statuses:
            status = statuses.popitem()[1]
        else:
            status = None

        print "  status: %r" % status
        self.store_presence(entity, False, None, status, 0)

    def error(self, failure):
        print failure

class RosterMonitor(Monitor):

    def connectionInitialized(self):
        self.send("<iq type='get'><query xmlns='jabber:iq:roster'/></iq>")
        Monitor.connectionInitialized(self)

    def subscribeReceived(self, entity):
        self.subscribed(entity)

        # return the favour
        self.subscribe(entity)

    #def subscribedReceived(self, entity):
    #    pass

    def unsubscribeReceived(self, entity):
        self.unsubscribed(entity)

        # return the favour
        self.unsubscribe(entity)

    def unsubscribedReceived(self, entity):
        d = self.storage.remove_presences(entity)
        d.addErrback(self.error)



# Copyright (c) 2005-2006 Ralph Meijer
# See LICENSE for details

"""
Feed retrieval and parsing.

This module implements asynchronous download of RSS and Atom feeds via
HTTP, that are subsequently parsed using the Universal Feed Parser.
"""

import copy
from planet import feedparser, scrub

from twisted.internet import defer, reactor
from twisted.web import client

feeds = ['http://test.ralphm.net/blog/atom']

feedparser.SANITIZE_HTML = 0
feedparser.RESOLVE_RELATIVE_URIS = 0

class NotModified(Exception):
    pass

class Headers(object):
    """
    Helper module to keep HTTP headers.
    """

    def __init__(self, headers):
        self.dict = dict([(k, v[0]) for k, v in headers.iteritems()])
        self.get = self.dict.get

    def getheader(self, header):
        return self.get(header.lower(), None)

class FeedResource(object):
    """
    Fake resource that behaves like a feed retrieved via HTTP.

    This object holds the URL and HTTP status and headers just like
    Universal Feed Parser's internal fetcher. It is used to make UFP
    have the same information available after having asynchronously
    retrieved the feed.
    """

    def __init__(self, data, url, status, headers):
        self.data = data
        self.url = url
        self.status = status
        self.headers = Headers(headers)

    def info(self):
        return self.headers

    def read(self):
        return self.data

class HTTPFeedGetter(client.HTTPPageGetter):
    """ HTTP page getter for feeds. """

    def handleResponse(self, response):
        # Store the response in cache
        re = self.headers.get("etag", None)
        rl = self.headers.get("last-modified", None)
        rd = self.headers.get("date", None)
        if re or rl or rd:
            cache = {}
            if re:
                cache['etag'] = re[-1]
            if rl and rd:
                cache['last-modified'] = rl[-1]
            elif rd and not rl:
                cache['last-modified'] = rd[-1]

            self.factory.cache[self.factory.original_url] = cache

        # Act like a normal HTTPPageGetter
        client.HTTPPageGetter.handleResponse(self, response)

    def handleStatus_301(self):
        """ Handle status 301: Moved Permanently. """
        client.HTTPPageGetter.handleStatus_301(self)
        self.factory.original_url = self.factory.url
        self.factory.real_status = self.status

    def handleStatus_302(self):
        client.HTTPPageGetter.handleStatus_301(self)
        self.factory.real_status = self.status

    def handleStatus_303(self):
        self.factory.method = 'GET'
        client.HTTPPageGetter.handleStatus_301(self)
        self.factory.real_status = self.status

    def handleStatus_304(self):
        """ Handle status 304: Not Modified. """
        self.factory.noPage(NotModified())

    handleStatus_307 = handleStatus_302


class HTTPClientFeedFactory(client.HTTPClientFactory):
    """
    Factory for retrieving feeds via HTTP and parsing it using the
    Universal Feed Parser.

    The factory keeps a cache, uses HTTP condition GETs (last-modified
    and etag) and asks for compression to minimize bandwith usage.
    """

    protocol = HTTPFeedGetter
    cache = {}

    def __init__(self, url, method='GET', postdata=None, headers=None,
                 agent="Twisted PageGetter", timeout=0, cookies=None,
                 followRedirect=1, useCache=1):

        self.original_url = copy.copy(url)
        self.real_status = None

        headers = headers or {}

        cached = useCache and self.cache.get(url, None)
        if cached:
            etag = cached.get('etag', None)
            last_modified = cached.get('last-modified', None)
            if etag:
                headers.setdefault('If-None-Match', etag)
            if last_modified:
                headers.setdefault('If-Modified-Since', last_modified)

        headers.setdefault('Accept-Encoding', 'gzip, deflate')
        headers.setdefault('Accept', feedparser.ACCEPT_HEADER)

        client.HTTPClientFactory.__init__(self, url=url, method=method,
                postdata=postdata, headers=headers, agent=agent,
                timeout=timeout, cookies=cookies, followRedirect=followRedirect)

    def page(self, page):
        if self.waiting:
            self.waiting = 0
            resource = FeedResource(page, self.url,
                                    self.real_status or self.status,
                                    self.response_headers)

            def doScrub(data):
                scrub.scrub(self.url, data)
                return data

            d = defer.maybeDeferred(feedparser.parse, resource)
            d.addCallback(doScrub)
            d.chainDeferred(self.deferred)

def getFeed(url, contextFactory=None, *args, **kwargs):
    """
    Download a web feed, keep a cache of already downloaded pages.

    Download a feed. Return a deferred, which will callback with a feed (parsed
    with the Universal Feed Parser) or errback with a description of the error.

    See HTTPClientCacheFactory to see what extra args can be passed.
    """
    scheme, host, port, path = client._parse(url)
    factory = HTTPClientFeedFactory(url, *args, **kwargs)
    if scheme == 'https':
        from twisted.internet import ssl
        if contextFactory is None:
            contextFactory = ssl.ClientContextFactory()
        reactor.connectSSL(host, port, factory, contextFactory)
    else:
        reactor.connectTCP(host, port, factory)
    return factory.deferred



#!/usr/bin/env python

# Copyright (c) 2005-2008 Ralph Meijer
# See LICENSE for details.

from setuptools import setup
from mimir import __version__

setup(name='mimir',
      version=__version__,
      description=u'Mimir daemons',
      author='Ralph Meijer',
      author_email='ralphm@ik.nu',
      url='http://mimir.ik.nu/',
      license='MIT',
      packages=[
          'mimir',
          'mimir.aggregator',
          'mimir.aggregator.test',
          'mimir.monitor',
          'twisted.plugins',
      ],
      package_data={'twisted.plugins': ['twisted/plugins/mimir_aggregator.py',
                                        'twisted/plugins/mimir_monitor.py']},
      data_files=[('share/mimir', ['db/monitor.sql'])],
      zip_safe=False,
      install_requires = ['wokkel >= 0.4.0',
                          'simplejson',
                          'FeedParser'],
)



# Copyright (c) 2005-2007 Ralph Meijer
# See LICENSE for details

"""
Create a aggregation service.
"""

from twisted.application import service
from twisted.python import usage
from twisted.words.protocols.jabber.jid import JID

from wokkel import component, pubsub
from wokkel.generic import FallbackHandler
from wokkel.iwokkel import IXMPPHandler

from mimir.aggregator import aggregator

class Options(usage.Options):
    optParameters = [
        ('feeds', None, 'feeds', 'Directory that holds the list of feeds'),
        ('jid', None, 'aggregator', 'JID of this component'),
        ('secret', None, 'secret', 'Secret to connect to upstream server'),
        ('rhost', None, '127.0.0.1', 'Upstream server address'),
        ('rport', None, '5347', 'Upstream server port'),
        ('service', None, None, 'Publish subscribe service JID'),
        ('web-port', None, None, 'Port to listen for HTTP interface service'),
    ]

    optFlags = [
        ('verbose', 'v', 'Show traffic'),
    ]

    def postOptions(self):
        try:
            self['rport'] = int(self['rport'])
        except ValueError:
            pass


def makeService(config):
    s = service.MultiService()

    # create XMPP external component
    cs = component.Component(config['rhost'], config['rport'],
                             config['jid'], config['secret'])
    cs.setServiceParent(s)

    # wait for no more than 15 minutes to try to reconnect
    cs.factory.maxDelay = 900

    if config["verbose"]:
        cs.logTraffic = True

    FallbackHandler().setHandlerParent(cs)

    # set up publish-subscribe client handler
    publisher = pubsub.PubSubClient()
    publisher.setHandlerParent(cs)

    # create aggregation service 
    storage = aggregator.FileFeedStorage(config['feeds'])
    ag = aggregator.AggregatorService(storage)
    ag.setServiceParent(s)

    # set up feed handler from publisher
    ag.handler = aggregator.IFeedHandler(publisher)
    ag.handler.service = JID(config['service'])

    # set up XMPP handler to interface with aggregator
    IXMPPHandler(ag).setHandlerParent(cs)

    # set up site to interface with aggregator

    from twisted.application import internet
    from twisted.web2 import channel, resource, server

    root = resource.Resource()
    root.child_setfeed = aggregator.AddFeedResource(ag)
    site = server.Site(root)
    w = internet.TCPServer(int(config['web-port']), channel.HTTPFactory(site))
    w.setServiceParent(s)

    return s



# Copyright (c) 2005-2008 Ralph Meijer
# See LICENSE for details

try:
    from twisted.application.service import ServiceMaker
except ImportError:
    from twisted.scripts.mktap import _tapHelper as ServiceMaker

mimirMonitor = ServiceMaker(
        "Mimir Monitor",
        "mimir.monitor.tap",
        "Mimir News and Presence Monitor",
        "monitor")



# Copyright (c) 2005-2007 Ralph Meijer
# See LICENSE for details

"""
Feed Aggregator.

Provides a service that connects to a Jabber server as a server-side component
and continuously aggregates a collection of Atom and RSS feeds, parses them and
republishes them via Jabber publish-subscribe.
"""

import re
import simplejson

from zope.interface import Interface, implements

from twisted.application import service
from twisted.internet import reactor, defer
from twisted.python import components, log
from twisted.python.filepath import FilePath
from twisted.words.protocols.jabber import xmlstream
from twisted.words.protocols.jabber.error import StanzaError
from twisted.words.xish import domish
from twisted.web import error
from twisted.web2 import http, resource, responsecode
from twisted.web2.stream import readStream

from wokkel import pubsub
from wokkel.iwokkel import IXMPPHandler
from wokkel.subprotocols import XMPPHandler

from mimir import __version__
from mimir.aggregator import fetcher, writer
from mimir.monitor.news import FeedParserEncoder

INTERVAL = 1800
NS_AGGREGATOR = 'http://mimir.ik.nu/protocol/aggregator'

RE_HANDLE = re.compile('^[-a-z0-9_]+$')

class InvalidHandleError(Exception):
    """
    Handle contains invalid characters.
    """


class IFeedHandler(Interface):
    """
    Handle aggregated feeds.
    """

    def entriesDiscovered(self, handle, feed, entries):
        """
        Called when new or updated entries have been discovered.

        @param handle: Handle for the feed the entries belong to.
        @type handle: C{unicode}
        @param feed: Output from the Universal Feed Parser.
        @type feed: L{feedparser.FeedParserDict}
        @param entries: List with entry details conforming to the
                        output of the Universal Feed Parser.
        @type entries: C{list}
        @rtype: L{defer.Deferred}
        """


class IAggregatorService(Interface):

    def setFeed(handle, url):
        """
        Associate a feed URL to a handle.

        If the handle is new, the feed is added to the aggregation schedule.

        @param handle: feed handle.
        @type handle: L{unicode}
        @param url: feed URL.
        @type url: L{str}
        """


class FileFeedStorage(object):
    def __init__(self, feedsDir):
        self.feedsDir = FilePath(feedsDir)
        self.feedListFile = self.feedsDir.child('feeds')
        self.feeds = None

    def _readFeedList(self):
        """
        Initialize list of feeds
        """

        fh = self.feedListFile.open()
        try:
            feedList = [line.split() for line in fh.readlines()]
        finally:
            fh.close()

        self.feeds = {}
        for handle, url in feedList:
            self.feeds[handle] = {'handle': handle,
                                  'href': url}

    def _writeFeedList(self):
        """
        Write out list of feeds
        """

        feedList = ["%s %s\n" % (f['handle'], f['href'])
                     for f in self.feeds.itervalues()]
        feedList.sort()

        fh = self.feedListFile.open('w')
        try:
            fh.writelines(feedList)
        finally:
            fh.close()

    def getFeedList(self):
        """
        Get the list of feeds.
        """
        if self.feeds is None:
            try:
                self._readFeedList()
            except:
                return defer.fail()

        return defer.succeed(self.feeds)

    def setFeedURL(self, handle, url):
        """
        Add or update the feed.
        """
        feed = {'handle': handle,
                 'href': url}

        self.feeds[handle] = feed

        try:
            self.storeFeed(feed)
            self._writeFeedList()
        except:
            return defer.fail()

        return defer.succeed(feed)

    def getFeed(self, handle):
        """
        Retrieve the feed as it was last aggregated.
        """

        try:
            feedFile = self.feedsDir.child("%s.feed.json" % handle)

            feed = self.feeds[handle]
            if feedFile.exists():
                fh = feedFile.open()
                try:
                    feed = simplejson.load(fh)
                except ValueError:
                    pass
                finally:
                    fh.close()
        except:
            return defer.fail()

        return defer.succeed(feed)

    def storeFeed(self, feed):
        """
        Write out a feed as a JSON file.
        """

        try:
            feedFile = self.feedsDir.child("%s.feed.json" % feed['handle'])
            if feedFile.exists():
                feedFile.moveTo(FilePath(feedFile.path + '.1'))
            fh = feedFile.open("w")
            try:
                simplejson.dump(feed, fh, indent=4, cls=FeedParserEncoder)
            finally:
                fh.close()
        except:
            return defer.fail()
        else:
            return defer.succeed(None)

class AggregatorService(service.Service):
    """
    Feed aggregator service.

    @ivar feedListFile: file that holds the list of feeds
    @type feedListFile: L{str}
    @ivar handler: handler of new and changed feed items.
    @type handler: object implementing L{IFeedHandler}
    """

    implements(IAggregatorService)

    agent = "MimirAggregator/%s (http://mimir.ik.nu/)" % __version__

    def __init__(self, storage):
        self.handler = None
        self.storage = storage

    def _callLater(self, *args, **kwargs):
        """
        Make callLater calls indirect for testing purposes.
        """

        return reactor.callLater(*args, **kwargs)

    def startService(self):
        """
        Read feeds from file and initialize service.
        """

        log.msg('Starting Aggregator')
        service.Service.startService(self)

        self.schedule = {}

        def scheduleFeeds(feeds):
            delay = 5
            for handle in feeds:
                self.reschedule(delay, handle)
                delay += 5

        d = self.storage.getFeedList()
        d.addCallback(scheduleFeeds)

    def stopService(self):
        log.msg('Stopping Aggregator')
        service.Service.stopService(self)

        calls = self.schedule.values()
        self.schedule = {}
        for call in calls:
            call.cancel()

    def setFeed(self, handle, url):
        if not RE_HANDLE.match(handle):
            return defer.fail(InvalidHandleError())

        try:
            self.schedule[handle].cancel()
        except KeyError:
            pass

        def scheduleFeed(result):
            if self.running:
                self.reschedule(0, handle, useCache=0)
            return result

        d = self.storage.setFeedURL(handle, url)
        d.addCallback(scheduleFeed)
        return d

    def aggregate(self, handle, useCache=1):
        def aggregateFeed(cachedFeed):
            headers = {}
            if useCache:
                etag = cachedFeed.get('etag', None)
                updated = cachedFeed.get('updated', None)
                if etag:
                    headers['If-None-Match'] = str(etag)
                if updated:
                    short_weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri',
                                      'Sat', 'Sun']
                    months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                              'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
                    headers['If-Modified-Since'] = \
                        '%s, %02d %s %04d %02d:%02d:%02d GMT' % (
                                short_weekdays[updated[6]],
                                updated[2],
                                months[updated[1] - 1],
                                updated[0],
                                updated[3], updated[4], updated[5])

            d = defer.maybeDeferred(fetcher.getFeed, str(cachedFeed['href']),
                                                     agent=self.agent,
                                                     headers=headers,
                                                     useCache=useCache)
            d.addCallback(self.workOnFeed, handle)
            d.addCallback(self.findFreshEntries, handle, cachedFeed)
            return d

        def setInterval(feed):
            if 'interval' not in feed:
                feed['interval'] = INTERVAL

        d = self.storage.getFeed(handle)
        d.addCallback(aggregateFeed)
        d.addCallback(setInterval)
        d.addCallback(self.updateCache)
        d.addErrback(self.notModified, handle)
        d.addErrback(self.logNoFeed, handle)
        d.addErrback(self.munchError, handle)
        d.addBoth(lambda feed: self.reschedule(feed['interval'], handle))
        return d

    def workOnFeed(self, result, handle):
        result['handle'] = handle

        if result.status == '301':
            log.msg("%s: Feed's location changed permanently to %s" %
                    (handle, result.href))
            self.storage.setFeedURL(handle, result.href)

        if result.feed:
            log.msg("%s: Got feed." % handle)
            if 'title' in result.feed:
                log.msg("%s: Title: %r " % (handle,
                                            result.feed.title))
        else:
            log.msg("%s: Not a valid feed." % handle)

        if result.bozo:
            log.err(result.bozo_exception, "%s: Bozo flag raised" % handle)

        for entry in result.entries:
            if 'id' not in entry and 'link' in entry:
                entry.id = entry.link

        return result

    def findFreshEntries(self, result, handle, cachedFeed):
        def add(entry, kind):
            log.msg("%s: Found %s entry" % (handle, kind))
            log.msg("%s:   id: %s" % (handle, entry.id))
            if 'title' in entry:
                log.msg("%s:   title (%s): %r" % (handle,
                                                  entry.title_detail.type,
                                                  entry.title_detail.value))
            discoveredEntries.append(entry)

        cacheIndexes = cachedFeed.get('indexes', {})

        newCacheIndexes = {}
        discoveredEntries = []

        index = len(result.entries)
        for entry in reversed(result.entries):
            index -= 1

            if 'id' not in entry:
                continue

            if entry.id not in cacheIndexes:
                add(entry, 'new')
            else:
                jsonEntry = simplejson.dumps(entry, cls=FeedParserEncoder)
                newEntry = simplejson.loads(jsonEntry)
                if cachedFeed['entries'][cacheIndexes[entry.id]] != newEntry:
                    add(entry, 'updated')

            newCacheIndexes[entry.id] = index

        result['indexes'] = newCacheIndexes

        if discoveredEntries:
            d = self.handler.entriesDiscovered(handle, result,
                                               discoveredEntries)
        else:
            d = defer.succeed(None)

        d.addCallback(lambda _: result)
        return d

    def updateCache(self, result):
        log.msg("%s: updating cache" % result["handle"])
        return self.storage.storeFeed(result)

    def notModified(self, failure, handle):
        failure.trap(fetcher.NotModified)
        log.msg("%s: Not Modified" % handle)

    def reschedule(self, delay, handle, useCache=1):
        def cb():
            del self.schedule[handle]
            self.aggregate(handle, useCache=useCache)

        self.schedule[handle] = self._callLater(delay, cb)

    def logNoFeed(self, failure, handle):
        failure.trap(error.Error)
        log.err(failure, "%s: No feed" % handle)

    def munchError(self, failure, handle):
        log.err(failure, "%s: Unhandled Error" % handle)


class XMPPControl(XMPPHandler):

    def __init__(self, service):
        self.service = service

    def connectionInitialized(self):
        xpath = '/iq[@type="set"]/aggregator[@xmlns="%s"]/feed' % NS_AGGREGATOR
        self.xmlstream.addObserver(xpath, self.onFeed, 0)

    def onFeed(self, iq):
        handle = str(iq.aggregator.feed.handle or '')
        url = str(iq.aggregator.feed.url or '')

        iq.handled = True

        def success(_):
            return xmlstream.toResponse(iq, 'result')

        def trapInvalidHandle(failure):
            failure.trap(InvalidHandleError)
            raise StanzaError('bad-request', text='Invalid handle')

        def error(failure):
            if failure.check(StanzaError):
                exc = failure.value
            else:
                log.err(failure)
                exc = StanzaError('internal-error')
            return exc.toResponse(iq)

        if handle and url:
            d = self.service.setFeed(handle, url)
            d.addCallback(success)
            d.addErrback(trapInvalidHandle)
        else:
            d = defer.fail(StanzaError('bad-request'))

        d.addErrback(error)
        d.addCallback(self.send)

components.registerAdapter(XMPPControl, IAggregatorService,
                                        IXMPPHandler)


class AddFeedResource(resource.Resource):
    """
    Resource to add a new feed to be aggregated.
    """

    def __init__(self, service):
        self.service = service

    http_GET = None

    def http_POST(self, request):
        def gotRequest(result):
            url = result['url']
            handle = result['handle']
            d = self.service.setFeed(handle, url)
            d.addCallback(lambda _: self.service.handler.checkNode(handle))
            return d

        def createResponse(result):
            response = {'uri': 'xmpp:%s?;node=%s' % (
                self.service.handler.service.full(),
                result)}
            return http.Response(responsecode.OK, stream=simplejson.dumps(response))

        def trapInvalidHandle(failure):
            failure.trap(InvalidHandleError)
            return http.StatusResponse(responsecode.BAD_REQUEST,
                                       """Invalid handle""")

        data = []
        d = readStream(request.stream, data.append)
        d.addCallback(lambda _: ''.join(data))
        d.addCallback(simplejson.loads)
        d.addCallback(gotRequest)
        d.addCallback(createResponse)
        d.addErrback(trapInvalidHandle)
        return d



class AtomPublisher(object):

    implements(IFeedHandler)

    def __init__(self, protocol):
        self.protocol = protocol
        self.service = None
        self.writer = writer.ReconstituteWriter()


    def _getNode(self, handle):
        return 'mimir/news/%s' % handle


    def entriesDiscovered(self, handle, feed, entries):
        log.msg("%s: publishing items" % handle)

        node = self._getNode(handle)

        items = []
        for entry in entries:
            try:
                entryDoc = self.writer.generate(feed, entry)
            except domish.ParserError:
                log.err(None, '%s: Error processing entry: %r' % (handle,
                                                                  entry.title))
            else:
                items.append(pubsub.Item(entry.id, entryDoc))

        if items:
            return self.protocol.publish(self.service, node, items)
        else:
            return defer.succeed(None)

    def checkNode(self, handle):
        def trapConflict(failure, node):
            failure.trap(StanzaError)
            exc = failure.value
            if exc.condition != 'conflict':
                return failure
            else:
                return node

        node = self._getNode(handle)
        d = self.protocol.createNode(self.service, node)
        d.addErrback(trapConflict, node)
        return d

components.registerAdapter(AtomPublisher, pubsub.IPubSubClient,
                                          IFeedHandler)
